package PGP5::Key;

use strict;
use Time::Local;                        
use Data::Dumper;
use PGP5::Pipe;

=head2 PGP5::Key

The C<PGP5::Key> object is used to store the individual key
information. It is primarily used by the C<PGP5::Keyring> object and
for passing to the various methods that accept key parameters to
encrypt and sign documents. 

Future revisions will provide actual methods to do key comparison for
the trust and validity factors. These methods will provide a
standardized way to determine which keys can be trusted and which
keys should not be used at all.

=cut 

=item * PGP5::Key::new

        $key = new PGP5::Key $keyline;

This is the constructor for the C<PGP5::Key> object. This is primarily
used by the C<PGP5::Keyring> methods. The C<PGP5::Keyring> methods keep
track of the keys and maintain the Trust and Validity components.
About the only useful method is the C<PGP5::Key::Fingerprint>, which
will return a string that is the finger print of the given key.

=cut

sub new
{
    my($class, $keyline, $keyring) = @_;
  
    chomp $keyline;
    my($bits, $keyid, $date, $expire, $algo, $use) = PGP5::Key->_keyparse ($keyline);
                            
    # $keyring ||= new PGP5::Keyring;
    my $self = { Keyring    =>      $keyring,
                Bits        =>      $bits,
                Keyid       =>      $keyid,
                Date        =>      $date,
                Expire      =>      $expire,
                Algorithm   =>      $algo,
                Use         =>      $use,
                Owner       =>      [],
                Email       =>      [],
                };

    bless $self, $class;
        
    # Add on the ID information to the key object
    # $self->Add_ID ($owner);
  
    return($self);
}


=item + PGP5::Key::Add_ID

        Add_ID $key $desc;

The C<Add_ID> method will add identification information to the owner 
and email portions of the given C<PGP5::Key> object. This is to support
keys that multiple identification packets associated with them.

=cut                   
                                          
sub Add_ID
{
    my ($self, $desc) = @_;
                                               
  # we have a total of three types of entries for the description
  #     full name <email@domain>        /\<.+\>/
  #     email@domain                    /[\w\.\-\+]@[\w\.\-\+]/
  #     full name                       all other 
  
    if ($desc =~ /\<.+\>/)
    {
        $desc =~ /([^\<]+)\s+\<(.+)\>/;
        push (@{$self->{Owner}}, $1);
        push (@{$self->{Email}}, $2);
    } elsif ($desc =~ /[\w\.\-\+]@[\w\.\-\+]/)
    {
        push (@{$self->{Owner}}, undef);
        push (@{$self->{Email}}, $desc);
    } else
    {
        push (@{$self->{Owner}}, $desc);
        push (@{$self->{Email}}, undef);
    }
}


=item * PGP5::Key::Add_Sig



=cut

sub Add_Sig
{
    my ($self, $line) = @_;
  
  
}


=item * PGP5::Key::Trust

This will set and/or retrieve the trust factor. Currently, this routine
will just store what is sent to it. Need to define some "trust" 
variables and provide useful routines to use them.

=cut


sub Trust
{
    my ($self, $trust) = @_;
  
    $self->{Trust} = $trust if ($trust);
    $self->{Trust};
}


=item * PGP5::Key::Validity

This function will set and/or return the validity factor. This 
subroutine is very much like PGP5::Key::Trust. It also needs to be
worked on quite a bit.

=cut
     

sub Validity
{
    my ($self, $validity) = @_;
  
    $self->{Validity} = $validity if ($validity);
    $self->{Validity};
}


=item * PGP5::Key::Fingerprint

        $fingerprint = Fingerprint $key;


=cut

# does the Fingerprint method belong in the key management stuff?

sub Fingerprint
{
    my ($self) = shift;
    my $pgp=$self->{Keyring} ? $self->{Keyring}->{PGP}
                             : new PGP5::Pipe;

    # what hackery!  Has ta be a better way
    $pgp->Exec('key', "-ll 0x$self->{Keyid}", *FIN, *FOUT, *FERR);
    my $fingerprint='';
    my $next=1;
    while (<FOUT>)
    {
        if(/^pub\s+\d+\s+$self->{Keyid}/)
        { $next = 1; }
        elsif($next and /^f(\d+)\s+Fingerprint\1\s+=\s+([0-9A-Z ]+)$/)
        {
            $fingerprint = $2;
            last;
        }
    }
  
    return $fingerprint;
}


=item * PGP5::Key::Format

        $formatted_text = Format $key %args;

This method will return a formatted text string for a key. It 
is essentially the same as do a 'pgp -kv' or 'pgp -kvv' for
a key object. Currently the only argument that C<Format> will 
recognize is the C<Verbose> argument. The C<Verbose> parameter
will list the signatures that have certified the current 
key object. 

=cut

sub Format
{
    my ($self, %args) = @_;
    my ($day, $month, $year) = $self->_date ($self->{Date});

    my $text = sprintf ("pub  %4d %s %4d-%02d-%02d %s %-15s %s\n",
                $self->{Bits}, $self->{Keyid},
                $year, ($month+0), ($day+0), 
                $self->{Expire},
                $self->{Algorithm},
                $self->{Use},);
                
                
    my $index = 0;
    while ($self->{Owner}->[$index] || $self->{Email}->[$index])
    {
        $text .= sprintf ("uid  %s\n", $self->_desc ($self->{Owner}->[$index],
                        $self->{Email}->[$index]));
        if (exists $args{Verbose})      # produce a list of signatures
        {
    
        }
        $index++;
    }
  
    if($self->{'Trust'} and $self->{'Validity'})
    {
        my $desc=$self->_desc($self->{Owner}->[0], $self->{Email}->[0]);
        $text.="\n  KeyID    Trust     Validity  User ID\n";
        $text.=sprintf('  %-8s %-9s %-9s %s',
                        $self->{'Keyid'}, $self->{'Trust'}, $self->{'Validity'},
                        "$desc\n");
    }

    return $text;
}


              
sub _keyparse
{
    my($self, $keyline) = @_;
                        
    my($bits, $keyid, $year, $mon, $day, $eyear, $emon, $eday, $algo, $use) =
        ($keyline =~ /^pub\s+       # magic marker
                      (\d+)[\/\s]+  # number of bits
                      (?:0x)?       # 5.0i beta didn't have this...
                      (\w+)\s+      # HEX key ID
                      (\d+)[-\/](\d+)[-\/](\d+)\s+          # created date
                      (\d+|----)[-\/](\d+|--)[-\/](\d+|--)  # expire date
                      (.+?)\s+      # Algorithm
                      (.+)$/x);     # Use
    die $keyline unless $bits;
    return unless defined $bits;

    my $date = &Time::Local::timegm (0, 0, 0, $day, $mon, $year-1900);
    my $expire = '----------';
    $expire = &Time::Local::timegm (0, 0, 0, $eday, $emon, $eyear-1900)
        if($eyear ne '----');
                                     
    return ($bits, $keyid, $date, $expire, $algo, $use);
}
                                 

sub _date
{     
    my $self = shift;
    my (@tm) = gmtime (shift);
    return ($tm[3], $tm[4], $tm[5]+1900);
}

                                    
sub _desc
{
    my ($self, $owner, $email) = @_;
  
    return ("$owner <$email>") if ($owner && $email);
    return ("$owner") if (!$email);
    return ("$email") if (!$owner);
    return ("*** No ID on key ***");
}


=head2 Known Bugs and Limitations

=item + Hopefully none, proabably many!

=head2 Author

        Gerard Hickey
        RR 2  Box 409
        Lower Main St.
        North Berwick, ME   03906
        hickey@ctron.com

=head2 Copyrights

        Copyleft (l) 1996, by Gerard Hickey

What this means is that this program may be copied freely given that
there is no payment in exchange for this program, and that all the
source is left intact with all comments and documentation. If you
wish to modify this program to correct bugs or to extend it's
usefullness, please coordinate such actions with the author.

=cut

1;
