
=head2 PGP5::Keyring

The C<PGP5::Keyring> object is used to perform key management functions.

=cut


package PGP5::Keyring;

use strict;
use Config;
use Carp;
use PGP5::Key;
use PGP5::Pipe;

# @PGP5::Keyring::ISA = qw(PGP5::Pipe);

=item * PGP5::Keyring::new

        $Keyring = new PGP5::Keyring $pgpkeyring;


=cut


sub new 
{   
    my ($class, $pgp) = @_;

    $pgp=new PGP5::Pipe $pgp unless ref $pgp;
    return 0 if $pgp==0;

    my $self = {
                PGP             =>  $pgp,
                Keys            =>      [],
                Modified        =>      1
           };

    bless $self, $class;
  
    # Need to update the Keys field so that it is useful.
    $self->List_Keys;
    return $self;
};


=item * PGP5::Keyring::Add_Key

        Add_Key $Keyring %args;

Add a signature to the keyring. At this point, there is no error 
checking or verification that the key has been added.

The C<%args> associative array may contain the following:

        Text            The value is the public key
        File            File where the public key is stored

=cut

sub Add_Key
{
    my ($self, %args) = @_;
                    
    # PGP does not seem to like to take a keyring from stdin.
    # must place everything in a temporary file for processing.
  
    if ($args{Text})
    {
        open (TEMP, ">/tmp/.pgp.$$");
        print TEMP $args{Text};
        close TEMP;
    } elsif($args{File})
    {                         
        open TIN, "$args{File}"
            or croak "Unable to open $args{File}: $!";
        open (TOUT, ">/tmp/.pgp.$$");
        while(<TIN>) {print TOUT;}
        close TIN;
        close TOUT;
    } else
    {
        carp("Must specify Text or File");
        return 0;
    }
    
    $self->{PGP}->Exec('key', "a /tmp/.pgp.$$", *FIN, *FOUT, *FERR);
      
#   # send the key to the PGP process
#   if ($args{Text})
#   { print FIN "$args{Text}\n" }
#   else
#   {
#       open (KEY, "<$args{File}");
#       print FIN <KEY>;
#       close KEY;
#   }
    close FIN;
    close FOUT;
    close FERR;
    wait();
    $self->{Modified}++;
}
      

=item * PGP5::Keyring::Remove_Key

        Remove_Key $Keyring $key;

Remove a signature from a keyring.   

=cut


sub Remove_Key
{ 
    my ($self, $key) = @_;
  
    $self->{PGP}->Exec('key', "r -f 0x$key->{Keyid}", *FIN, *FOUT, *FERR);
    close FIN;
    close FOUT;
    close FERR;
    wait();
    $self->{Modified}++;
}


=item * PGP5::Keyring::Extract_Key

        $key = Extract_Key $Keyring $keyobj;

Extract a key from the specified keyring. A real simple dirty way of 
extracting the key.

=cut
 

sub Extract_Key
{
    my ($self, %args) = @_;
  
    $self->{PGP}->Exec ('key', "xa -f 0x$args{Keyid}", *FIN, *FOUT, *FERR);
  
    my @key = <FOUT>;
    close FOUT;
    close FIN;
    close FERR;
    wait();
    return (join ('', @key));
}


=item * PGP5::Keyring::Sign_Key

        Sign_Key $Keyring %args;

This method will sign a designated key with the 


=cut

sub Sign_Key
{
    my ($self, %args) = @_;
  
    # We absolutely need a password to continue!
    return if (! exists $args{Password});
  
    my $result = $self->{PGP}->Exec ('keys', "s 0x$args{Keyid}", *FIN, *FOUT, *FERR, 1);
  
    open (KEYB, ">/dev/tty");
                                       
    # now we get to act as a user to get the key signed!
    my $output;
    while (defined($output = <FOUT>))
    {
        PGP5::Pipe::Debug ("FOUT: $output");
        last if ($output =~ /user ID/ );
    }
    
    print KEYB "y\n";     # say yes it is the key we want
    print "Sent a 'y' keystroke...";
    
    while (defined($output = <FOUT>))
    {
        PGP5::Pipe::Debug ("FOUT: $output");
        last if ($output =~ /Enter pass phrase:/);
    }
    
    print KEYB "$args{Password}\n";
  
    # right now, we just hope that it signed it fine!
    close (FIN);
    close (FOUT);
    close (FERR);
    wait();
    # Keyring has been modified.
    $self->{Modified}++;
  
    $self->Extract_Key (Keyid => $args{Keyid});
}


=item * PGP5::Keyring::Generate_Key

        Generate_Key $Keyring;

Generate a new secret and public key set. This routine will not be
present in the first rev of code. It is also subject to change.

=cut
 

sub Generate_Key
{
    my ($self) = shift;

# be sure to use +nomanual as an option
    $self->{Modified}++;
}


=item * PGP5::Keyring::Revoke_Key

        $certificate = Revoke_Key $Keyring $Keyobj;

Produce a revocation certificate for the given key. Revocation is
actually a two step process. We must first mark the key as revoked.
This is the same as the C<Remove_Key> method. After flaging the key,
the key must be extracted to produce a revocation certificate.

=cut

sub Revoke_Key
{
    my ($self, $key) = @_;
                                                              
    $self->Remove_Key ($key);
    return ($self->Extract_Key ($key));
}

             
=item * PGP5::Keyring::List_Keys

        @{$keyobj} = List_Keys $Keyring;

List the keys on a given keyring. This routine simply captures the output
of the command C<pgp -kc $keyring> and does a quick parse on it. It
takes the lines that it parses, and constructs L<PGP5::Key> objects.
In the near future, this function will also pass the trust factors to the 
PGP5::Key object. We got it in the output, so why not use it.

=cut
 

sub List_Keys
{
    my ($self) = @_;
    my ($keyid, $trust, $validity, $desc);

    # do not call PGP if the keys have not been modified
    if (!$self->{Modified})
    {
        return (wantarray ? @{$self->{Keys}} : $self->{Keys});
    }

    # clear the old array and get a list of all the keys again
    $self->{Keys} = undef;
    $self->{PGP}->Exec('key', "-c", *FIN, *FOUT, *FERR);

    my $t=<FERR>;           # DAMN!  Why the fsck does FOUT end up in FERR?
    if($t and $t=~/Type\s+Bits/)
    {
        *FOUT=*FERR;
    }
    while (<FOUT>)
    {
        # warn "Read $_";
        # public key entry
        if(/^pub/)
        { push @{$self->{Keys}}, PGP5::Key->new($_, $self);
        } elsif(/^sig/)
        { $self->{Keys}->[$#{$self->{Keys}}]->Add_Sig($_);
        } elsif(/^\s{30,32}(.+)$/ or /^uid\s+(.+)$/)
        { $self->{Keys}->[$#{$self->{Keys}}]->Add_ID($1);
        }

        # public key trust entries follow
        last if (/^\s+KeyID\s+Trust\s+Validity\s+User ID/);
    }
    while (<FOUT>)
    {
        # print "Read '$_'";
        # valid entry?
        if(/^..(?:0x)?(\w+)\s+(\w+)\s+(\w+)\s+(.+)/)
        {
            $keyid = $1; $trust = $2; $validity = $3; $desc = $4;
            my $key = $self->Find(Keyid => $keyid);
            if($key)
            {
                ## Key doesn't exist... is it a private key?
             #   die "Can't find $keyid to trust it\n", Dumper $self unless($key);
                $key->Trust($trust);
                $key->Validity($validity);
            }
        }
    }

    # Now that we have the latest keyring data, reset the modified flag
    $self->{Modified}=0;
    # die Dumper($self);
    close FIN;
    close FOUT;
    close FERR;
    wait();
    return (wantarray ? @{$self->{Keys}} : $self->{Keys});
}


=item * PGP5::Keyring::Find

        @keys = Find $keyring %criteria;
        \@keys = Find $keyring %criteria;
        $key = Find $keyring %criteria; (Single match)

Function to locate a keys matching some criteria. This is not 
implemented as nicely as it should be (read kludge). The 
C<%criteria> array is used to specify what keys are to be selected.
The keys for the C<%criteria> array are as follows:

        Keyid           Key with specifed keyid
        Owner           Name of the owner of the key
        Email           Email address of owner
        Bits            Size of the key in bits
        Date            Date that the key was generated
        Desc            Owner and Email keys combined

The values for each specifed key (assocative array) are compared
using a case-insensitive regular expression. This means that 
only a portion of the key data needs to be specified to have it 
selected. This also means that specifing too little criteria
can cause several keys to be selected.

=cut 


sub Find
{
    my ($self, %criteria) = @_;
    my ($key, @match, $crit, $thing);

    NONMATCH:
    foreach $key (@{$self->{Keys}})
    {
        foreach $crit (keys %criteria)
        {
            next unless($key->{$crit} or $crit eq 'Desc');
            $thing=$criteria{$crit};
            if ($crit eq 'Desc')
            {
                for ($[ .. $#{$key->{Owner}})
                {
                    my $keydesc = "$key->{Owner}->[$_] $key->{Email}->[$_]";
                    $keydesc=~ tr/a-zA-Z0-9@.!\-//cd;
                    $keydesc =~ tr/       / /s;

                    my $desc = $thing;
                    # update the tr/// above too!
                    $desc =~ tr/a-zA-Z0-9@.!\-//cd;
                    $desc =~ tr/  / /s;

                    next NONMATCH if ($keydesc !~ /$desc/i);
                }
            } elsif (ref ($key->{$crit}) ne 'ARRAY')
            {
                next NONMATCH if ($key->{$crit} !~ /$thing/i);
            } else
            {
                foreach (@{$key->{$crit}})
                {
                    if($_ and /$thing/i)
                    {
                        push @match, $key;
                        next NONMATCH;
                    }
                    # next NONMATCH unless($_ and /$criteria{$crit}/i);
                }
                next NONMATCH;
            }
        }
        push @match, $key;
    }

    # return a scalar if there is only one match
    return '' if($#match==-1);
    return ($match[$[]) if ($#match == 0);
    return (wantarray ? @match : \@match);
}  
             



1;
