package PGP5::Pipe;

require 5.000;
use strict;

use English;
use Carp;
use File::Basename;
use IPC::Open3;
use Time::Local;
use Data::Dumper;
use User::pwent;

=over 4

=head1 NAME

PGP - perl module to work with PGP messages

=head1 SYNOPSIS

use PGP;

$message = new PGP $pgppath;

=head1 DESCRIPTION

The PGP module allow a perl script to work with PGP related files.

=cut

# $Log: Pipe.pm,v $
# Revision 0.3  1996/08/14 19:04:21  hickey
# + moved module to PGP5::Pipe to prevent conflicts
# + upgraded to Data::Dumper
# + added the Sign_Key method to PGP5::Keyring (not debugged yet)
# + PGP5::Pipe::Exec places filehandles in caller's package
#
# Revision 0.2  1996/01/27  15:40:57  hickey
# + PGP5::Keyring and PGP5::Key now inherits PGP object
# + PGP5::Keyring::Find now correctly works (filter on anything)
# + Timestamps are now correctly reported back to caller
# + Activated %r (path to keyring) in the PGP5::Exec_PGP method
# + Added support for multiple ID keys. (PGP5::Key::Add_ID)
#
# Revision 0.1  1996/01/10  02:22:18  hickey
# Initial alpha release
#

use vars qw($VERSION $RCSID);
$VERSION = '0.3';
$RCSID   = '$Id: Pipe.pm,v 0.3 1996/08/14 19:04:21 hickey Exp hickey $';
$PGP5::Pipe::debug=0;

=item * PGP5::new

	$pgp = new PGP [$pgppath], [$pgpexec];

Create the PGP encapsulation object. The standard location for the
PGP executable is /usr/local/bin/pgp.

=cut

sub new
{
    my $class = shift;
    my $pgppath = shift;
    my $pgpexec = shift || find_exec("pgp");

    unless($pgppath)
    {
        if($ENV{HOME})
        {   $pgppath="$ENV{HOME}/.pgp"; }
        else
        {   my $pw=getpwuid($UID);
            die "Unable to get user info for $UID" unless ref $pw;
            $pgppath=join '/', $pw->dir, '.pgp';
        }
    }
    my $realpath='';
    my @lookin=($pgppath, qw(/usr/lib/pgp /usr/local/lib/pgp));
    foreach my $path (@lookin)
    {
        if(-e "$path/pgp.cfg")
        {   $realpath=$path;
            last;
        }
    }
    unless($realpath)
    {
        carp 'PGP configuration file not found, looked in ('.join(', ', @lookin).')';
        return;
    }

    my $self = {    PGPPATH     =>      $realpath,
                    PGPe        =>      find_exec("${pgpexec}e"),
                    PGPs        =>      find_exec("${pgpexec}s"),
                    PGPv        =>      find_exec("${pgpexec}v"),
                    PGPk        =>      find_exec("${pgpexec}k"),
                    # PGPo        =>      find_exec("${pgpexec}o"),
                    # pgpo would have be VERY useful
            };
    $ENV{PGPPATH} = $realpath;

    bless $self, $class;
}

# The following function was suggested by Alva Couch. Can anyone
# think why they would call it rather than the new() method?

# projector function eliminates all non-PGP data and 
# returns a 'pure' PGP instance. 

sub PGP5
{
  my $self = shift;
  bless {	PGPPATH	=> $self->{PGPPATH}, 
		PGPexec => $self->{PGPexec}, 
        }, 'PGP5';
}


sub Debug
{
    my (@args) = @_;

    return unless $PGP5::Pipe::debug;
    warn join '', @args, "\n";
}


=item * PGP5::Exec

	$pid = Exec $pgp $args, $in, $out, $err, $nobatchmode;

Execute the PGP command and attach the C<$in>, C<$out>, C<$err> file handles. 
This should be fine for the moment, but need to look into making
sure that data is not written to a temporary file anywhere. The C<$nobatchmode>
parameter causes the PGP command to be executed without the +batchmode
parameter. This seems to only be necessary when a key is being signed.

The $args variable can have several substituted strings:

	%p	PGP path variable
	%r	Path to PGP keyring
	%k	Specified user

B<Note:> The above substitutions may change at any time. It is not
advised that you write applications with substitutions. Almost
certainly, the next release will not include substitutions.

The file handle variables--C<$in>, C<$out> and C<$err>--are send as
normal filehandle names, but they reside in the PGP package. For
example, the following procedure call is made:

        PGP->Exec ($args, *FIN, *FOUT, *FERR);

Even though the file handles were specified as C<FIN>, C<FOUT> and
C<FERR>; they must be referred to as C<PGP5::FIN>, C<PGP5::FOUT> and
C<PGP5::FERR> in the orignal procedure that made the call.

=cut


sub Exec
{
    my ($self, $type, $args, $fin, $fout, $ferr, $nobatchmode) = @_;
    my ($pgppath, $pgpcmd, $baseopts);

    if ($nobatchmode)
    {   $baseopts = '+force +verbose=1' }
    else
    {   $baseopts = '+force +batchmode +verbose=0' };

    # Variable substitutions
    $args =~ s/%p/$self->{PGPPATH}/g;
    $args =~ s/%r/$self->{PGPPATH}\/$self->{Keyring}/g;   # PGP5::Keyring
    $args =~ s/%k/0x$self->{Keyid}/g;                     # PGP5::Key

    # Put the file descriptors in the callers package
    # my ($fin, $fout, $ferr);
    # $fin = (caller)[0] . "::$in" unless ref $fin;
    # $fout = (caller)[0] . "::$out" unless ref $fout;
    # $ferr = (caller)[0] . "::$err" unless ref $ferr;

    $args.=' 2>&1' unless $ferr;
    my $exec=$self->{'PGP'.substr($type, 0, 1)};

    Debug ("PGP5::Exec=$exec $baseopts $args");

    # just to make sure that PGPPATH is exported!
    $ENV{PGPPATH} = $self->{PGPPATH};
#    $SIG{PIPE}=sub{ die join '', <$ferr>; } if $ferr;

    my @exec=($fin, $fout, $ferr, $exec, split /\s+/, "$baseopts $args");
    # warn "Exec: ", join "\n", @exec;
    
    my $result = open3(@exec) or
        croak "PGP5 @exec command error: $!";

    # die join '', <$ferr>;
    return $result;
}


=item * PGP5::Sign

	$signed_document = Sign $pgp %args;

The C<Sign> procedure will take a file or data and sign with a PGP
secret key. The default behavior is to sign the data with the last
secret key added to the keyring, but that can be overridden with the
I<Key> argument. This method always returns the signed document.

The C<%args> consist of a series of keys and values. Since there are
several variations in the way data can be signed, not all the
following options must be specified. This approach also makes it much
easier to scale to new versions of PGP with more options.

	Armor		The output should be ASCII armored
	Clear		Produce a "clear" signature
	Encrypt		Encrypt the resulting signed document with
			the given keyobj
	Detach		Create a detached signature
	File		Sign the specified file
	Key		Sign with the specified key object
	Nosave		Do not allow user to save message
	Password	The password to use for signing
	Signfile	The filename of the signed document
	Text		Data to be signed.
	Wipe		Remove the orignal file

The only absolute argument that is always required is the C<Password>. 

B<Examples>

 Sign $pgp Password => 'xyz', File => '/etc/motd', Clear => 1, Armor => 1;

This would return a signed copy of the F</etc/motd> file. In this
case, we use a file as the input, but the output is returned at the
method's termination. The orignal file remains in the clear, and the
signature is ASCII armored (Base64). 

 Sign $pgp Password => 'abc', Text => 'Important info', Armor => 1,
           Signfile => 'signed.asc', Key => $keyobj;

This is sort of the reverse of the first example. It takes what is in
the C<Text> field and signs it. It then puts the result in the file
F<signed.asc> and returns it to the caller. In this case, the entire
message is ASCII armored including the orignal text (i.e. C<Text>).
We also specify another secret key to produce the signature. For more
information on the the key objects, please see L<"PGP5::Key"> section.

=cut


sub Sign
{
    my ($self, %args) = @_;
    my ($options, $key, $document);

    # Debug ("PGP5::Sign Args=", Dumper \%args);

    $options = '-f';
    $options .= 'a' if ($args{Armor} == 1);
    $options .= 'b' if ($args{Detach} == 1);
    $options .= 't' if (exists $args{Clear});
    $options .= 'w' if ($args{Wipe} == 1);
    $options .= 'm' if ($args{Nosave} == 1);

    # setup of encryption if we are doing any
    if (defined $args{Encrypt})
    {
        $options .= 'e';
        foreach $key (@{$args{Encrypt}})
        {
            $options .= " 0x$key->{Keyid}";
        }
    }

    # When signing a document, we always have a password.
    $options .= " -z $args{Password}";

    Debug ("PGP5::Sign Options=$options");

    # procede to send the document to PGP.
    $self->Exec('sign', $options, *FIN, *FOUT, *FERR);

    if ($args{File})
    {
        open (PLAIN, "< $args{File}")
            or carp "$args{File} not found";
        print FIN <PLAIN>;
        close (PLAIN);
    } else
    {
        print FIN $args{Text};
    }
    close (FIN);

    $document = join ('', (<FOUT>));
    close FOUT;
    close FERR;
    wait();

    if ($args{Signfile})
    {
        open (SIGN, "> $args{Signfile}")
            or carp "Can not create $args{Signfile}";
        print SIGN $document;
        close (SIGN);
    }

    return ($document);
}


=item * PGP5::Encrypt

	$encrypted_document = Encrypt $pgp %args;

The C<Encrypt> method produces an encrypted document with the given
public keys specified by C<Key>. The C<Encrypt> method follow the
same conventions as the C<Sign> method. The data to be encrypted can
be sent to the method or can reside in a file. The resulting
encrypted data can also reside in a file or be sent back to the caller. 

In addition to encrypting a document, the document can also be signed
by using the C<Sign> key in the C<%args> array. If the document is to
be signed by the default secret key (last key added to the secret
keyring), then C<Sign> can be left undefined or contain something
other than a reference to a key object. Otherwise the C<Sign> key
should contain a reference to a specific key object (see
L<"PGP5::Key">).

	Armor		The output should be ASCII armored
	Encryptfile	The filename of the encrypted document
	File		Encrypt the specified file
	Key		Encrypt with the specified key object
	Nosave		Do not allow user to save message
	Password	The password to use for signing
	Sign		In addition to encrypting, sign the document
	Text		Data to be encrypted
	Wipe		Remove orignal file

=cut


sub Encrypt
{
    my($self, %args) = @_;
    my($options, $document, $key);
  		    
    # Debug ("PGP5::Encrypt Args=", Dumper \%args);

    $options = '-f';
    $options .= 'a' if ($args{'Armor'});
    $options .= 's' if (exists $args{'Sign'});
    $options .= 'w' if ($args{'Wipe'});
    $options .= 'm' if ($args{'Nosave'});
    $options .= 't' if ($args{'Ascii'});

    # process the Key variable
    if (ref $args{'Key'} eq 'ARRAY')
    {
        foreach $key (@{$args{Key}})
        {
            $options .= " -r 0x$key->{Keyid}";
        }
    } elsif(ref $args{'Key'})
    {
        $key=$args{'Key'};
        $options .= " -r 0x$key->{Keyid}";
    } else
    {
        $options .= " -r $args{Key}";
    }

    # If we are also signing, we need to tell which key and password.
    $options .= " -u 0x$args{'Sign'}->{'Keyid'}" if (defined $args{'Sign'}->{'Keyid'});
    $options .= " -z '$args{'Password'}'" if (defined $args{'Password'});

    Debug ("PGP5::Encrypt Options=$options");

    # procede to send the document to PGP.
    # FOUT and FERR are switched because pgpe seems to FUCK around
    $self->Exec('encrypt', $options, *FIN, *FOUT, *FERR);

    if ($args{'File'})
    {
        open (PLAIN, "< $args{'File'}")
            or carp "Can't open $args{'File'}: $!";
        while(<PLAIN>) { print FIN; }
        close (PLAIN);
    } elsif($args{'Text'})
    {
        # Debug("Text is $args{Text}");
        print FIN $args{'Text'};
    }
    # autoflush FIN ();
    print FIN sprintf("\%c", 4);
    close (FIN);

    $document = join ('', <FOUT>);
    my $t=join '', <FERR>;
    close FOUT;
    close FERR;
    wait();

    # return $document;
    if ($args{'Encryptfile'})
    {
        open (ENCRYPT, "> $args{'Encryptfile'}")
            or carp "Can not create $args{'Encryptfile'}: $!";
        print ENCRYPT $document;
        close (ENCRYPT);
    }
    my($t1, $t2);
    if($document=~/(.*)   # farting!
                   (-{5}BEGIN\s+ PGP\s+ MESSAGE-{5}
                    .+
                    -{5}END\s+ PGP\s+ MESSAGE-{5})
                   (.*)   # after maybe errors?
                   /xs)
    {   $t1=$1; $document=$2; $t2=$3;
    } elsif($t=~/(.*)   # farting!
                   (-{5}BEGIN\s+ PGP\s+ MESSAGE-{5}
                    .+
                    -{5}END\s+ PGP\s+ MESSAGE-{5})
                   (.*)   # after maybe errors?
                   /xs)
    {   $t1=$1; $document=$2; $t2=$3;
    } else
    {
        return("ERROR: $document\n$t\n");
    }

    return ($document) unless wantarray;
    # die ($document, "$t1\n$t2");
    return ($document, "ERROR: $t1\n$t2 : ERROR");
}


=item * PGP5::Decrypt

	\%stats = Decrypt $pgp %args;

C<Decrypt> will use a PGP secret key to decrypt a message. The secret
key must reside on the secret keyring. The C<Decrypt> method follows
the same conventions for data transfer that C<Sign> and C<Encrypt>
follow. The resulting associative array that is sent back contains
three fields:

	Text		The decrypted document
        Signature       PGP5::Key object of the signer (if any)
	Time		Time document was signed (if any)
        Key             PGP5::Key object used to decrypt document

The following are the accepted arguments:

	Password	Password to use for decrypting
	File		File to decrypt
	Keyring		Needed to return info about document
	Plainfile	File to put the data in
	Text		Document to decrypt
	Wipe		Remove original file

The C<Password> argument is required to perform the decryption of the
document. The C<Keyring> argument is also required if any document 
information is to be returned.

=cut


sub Decrypt
{
    my($self, %args) = @_;
    my($options, $document, $key, @keys);
  		    
    # Debug ("PGP5::Decrypt Args=", Dumper \%args);

    $options = "-f ";
    $options = "-z '$args{Password}'" if (defined $args{Password});

    Debug ("PGP5::Decrypt Options=$options");

    # procede to send the document to PGP.
    $self->Exec ('decrypt', $options, *FIN, *FOUT, *FERR);

    if (defined $args{File})
    {
        open (ENCRYPT, "< $args{File}") || carp "$args{File} not found";
        while(<PLAIN>) {   print FIN; }
        close (ENCRYPT);
    } else
    {
        print FIN $args{Text};
    }
    close (FIN);

    $document = join ('', <FOUT>);

    if ($args{Plainfile})
    {
        open (PLAIN, "> $args{Plainfile}") || carp "Can not create $args{Plainfile}";
        print PLAIN $document;
        close (PLAIN);
    }

    if (defined $args{Keyring})
    {
        my $keyring = $args{Keyring};
    
        my $signature='';
        my $time='';
        my $key='';
        # gather stats on the decrypted document
        while (<FERR>)
        {
            # Encryption fields
            if(/Key ID (\w+)\,/i)
            { $key = Find $keyring  Keyid => $1 }
	
            # Signature fields
            if(/^Good signature from user "(.+)"/i)
            { $signature = Find $keyring  Desc => $1 }
            if(/^Signature made (\d+)\/(\d+)\/(\d+) (\d+):(\d+)/)
            { $time = &timegm (0, $5, $4, $3, $2-1, ($1 > 1900) ? $1 - 1900 : $1) }
        }
        close FOUT;
        close FERR;
        wait();
        return ({
		Text 		=>	$document,
  	     	Signature	=>	$signature,
	     	Time     	=>	$time,
	     	Key      	=>	$key
	    });
    } else
    {
        close FOUT;
        close FERR;
        wait();
        return ( { Text     =>      $document   } );
    }
}


=item * PGP5::Info

	\%doc = Info $pgp %args;

C<Info> returns an associative array or a reference to an
associative array to the caller. This returned structure contains
information about the document that is sent to the C<Info>
method. The returned structure is fairly straight forward:

	Text		The decrypted document
        Signature       PGP5::Key object of the signer (if any)
	Time		Time document was signed (if any)
        Key             PGP5::Key object used to decrypt document

The C<Info> method currently accepts the following arguments:

	File		File to decrypt
	Text		Document to decrypt
	
At this point, we cheat with the C<Info> method. Basically
we send the document through the C<Decrypt> method and grab the
results.

=cut


sub Info
{
    my ($self, %args) = @_;
    my $info = $self->Decrypt (%args, Plainfile => '/dev/null');
    return ($info);
}

sub find_exec
{
    my $exe = shift;

    # Debug("Looking for $exe");
    return $exe if $exe eq 'test';
    foreach my $cmd (split /;/, $exe)
    {
    	$cmd =~ s/^\s+//;

	    # remove any options
    	my $name = ($cmd =~ /^(\S+)/)[0];

	    # check for absolute or relative path
    	return ($cmd)
	        if (-x $name and ! -d $name and $name =~ m:/:);

    	$cmd=$1 if($cmd=~m!^.+[\\/]([^\\/]+)$!);
    	$name = ($cmd =~ /^(\S+)/)[0];

        # Debug("Scaning for $name");

        my $sep=':';
        $sep=';' if($^O eq 'MSWin32');
        foreach my $dir (split(/$sep/, $ENV{PATH}))
        {
            if (-x "$dir/$name" && ! -d "$dir/$name")
            {
    	        return "$dir/$cmd";
            }
            # Debug("Looked at $dir/$name");
    	}
    }
    return '';
}

1;
